
import math
from typing import Tuple, overload
from .types import Vector2


def clamp(value, minimum, maximum):
    """
    Clamps a value between a minimum and maximum range
    """
    return max(minimum, min(maximum, value))

def lerp(start: float, end: float, time: float, duration: float) -> float:
    """
    Linearly interpolate between two values over a given amount of time
    
    ### Params
    - `start`: start value
    - `end`: end value
    - `time`: current time
    - `duration`: duration of the interpolation
    
    ### Return
    interpolated value
    """
    return start + (end - start) * (time / duration)


@overload
def lerp(start:float, stop:float, amount:float)->float:
    return start + (stop-start)*amount


def map_range(value: float, start1: float, stop1: float, start2: float, stop2: float) -> float:
    """
    Maps a value from one range to another
    
    ### Params
    - `value`: value to be mapped
    - `start1`: start of the first range
    - `stop1`: end of the first range
    - `start2`: start of the second range
    - `stop2`: end of the second range
    
    ### Return
    mapped value
    """
    return (value - start1) / (stop1 - start1) * (stop2 - start2) + start2


def normalize(vector: Tuple[float, float] | Vector2) -> Vector2:
    """
    Normalize a 2D vector
    
    ### Params
    - `vector`: vector to be normalized
    
    ### Return
    Normalized vector
    """
    x, y = vector
    length = math.sqrt(x**2 + y**2)
    return x/length, y/length

def distance(p1: Tuple[float, float] | Vector2, p2: Tuple[float, float] | Vector2) -> float:
    """
    Calculates distance between two points
    
    ### Params
    - `p1`: first point
    - `p2`: second point
    
    ### Return
    distance
    """
    x1, y1 = p1
    x2, y2 = p2
    return math.sqrt((x1-x2)**2 + (y1-y2)**2)



"""

"""

from ezsgame.types import Pos, Size

class World:
    """
    #### Defines the "world"
    The `pos` and `size` defined the "view" of the world, objects outside of the view won't be drawn.
    """

    pos = Pos(0, 0)
    size = Size(0, 0) # will be set on window post init

    @classmethod
    def is_inside(cls, obj) -> bool:
        """
        #### Returns `True` if `obj` is inside the world
        """
        return cls.pos[0] <= obj.pos[0] <= cls.pos[0] + cls.size[0] and cls.pos[1] <= obj.pos[1] <= cls.pos[1] + cls.size[1]
        


from abc import ABC, abstractmethod
from enum import Enum
from .global_data import DATA

EventHandler = DATA.EventHandler

class ComponentGroup:
    """
	Handlers componets as a group, mainly the Mount/Unmount of each components
    """
    
    def __init__(self, object, *components):
        self.object = object
        self.components = {}

        if components:
            self.add(*components)

    def __del__(self):
        self.clear()
        del self

    def __str__(self):
        t = ", ".join([*map(str, self.components.values())])
        return f"<Component Group : [{t}] >"

    def __repr__(self):
        return self.__str__()

    def clear(self):
        self.remove(*self.components.keys())

    def remove(self, *components):
        for component in components:
            comp = self.components.get(component, None)

            if comp:
                comp.__remove()
                del self.components[component]

    def add(self, *components):
        for comp in components:
            comp_name = eval(f"{type(comp).__name__}")

            if comp and comp_name not in self.components:
                self.components[comp_name] = comp(self.object)

    def __contains__(self, thing):
        return eval(f"{type(thing).__name__}") in self.components.keys()

    def toggle(self, component):
        if eval(f"{type(component).__name__}") in self.components:
            self.remove(component)
            return False
        else:
            self.add(component)
            return True

    def __getitem__(self, other):
        if isinstance(other, int):
            return self.components[other]
        elif isinstance(other, slice):
            return self.__getslice__(other)

        else:
            if other in self.components:
                return self.components[other]
            else:
                raise KeyError(
                    f"Cannot get Component <{other}>  because not found in {self}")

    def __delitem__(self, other):
        if isinstance(other, int):
            del self.components[list(self.components.keys())[other]]

        elif isinstance(other, slice):
            self.__delslice__(other)

        else:
            if other in self.components:
                self.remove(other)
            else:
                raise KeyError(
                    f"Cannot delete Component <{other}>  because not found in {self}")

    def __getslice__(self, other):
        return [*self.components.values()][other.start:other.stop:other.step]

    def __get_components_keys(self, other):
        return [*self.components.keys()][other.start:other.stop:other.step]

    def __delslice__(self, other):
        for comp in self.__get_components_keys(other):
            self.remove(comp)

    def __get_key_by_index(self, index):
        return list(self.components.keys())[index]

    def __setitem__(self, other, item):
        if item not in self.components:
            try:
                comp = self.__get_key_by_index(other)
                self.remove(comp)
                self.add(item)
            except:
                raise TypeError("ComponentGroup can only contain Components")

        else:
            del self.components[other]

    def __len__(self):
        return len(self.components)

    def __iter__(self):
        self.__current_index = 0
        return iter(self.components.values())

    def __next__(self):
        if self.__current_index >= len(self.components):
            raise StopIteration
        else:
            self.__current_index += 1
            return self.components.values()[self.__current_index - 1]

class Component(ABC):
    def __init__(self):
        self.name = self.__class__.__name__

    def __str__(self):
        return f"<Component : {self.name}>"

    def __repr__(self):
        return self.__str__()

    def __remove(self):
        del self

    @abstractmethod
    def init(self, obj):
        pass

    @abstractmethod
    def call(self, obj):
        pass

    @abstractmethod
    def activate(self, obj):
        pass

    @abstractmethod
    def deactivate(self, obj):
        pass

    @abstractmethod
    def remove(self, obj):
        pass

    



from pathlib import Path
from ..components import Component
from ..global_data import get_window
from ..objects.object import Object
from typing import Any, Dict, Iterable
from ..styles.style import Styles
import pygame as pg
from ..styles.units import Measure
from ..types import Pos, Size


class Image(Object):
    r"""
    #### Image
    #### Parameters
    - `pos`: position of the image `[x, y]`
    - `size` : size of the image `[width, height]`
    - `image` : path to image file `str`

    #### Optional Arguments
    - `scale` : scale the image to the size `bool`
    - `components` : components to add in the object `[Component, ..]`
    - `styles` : Styles
    """

    def __init__(
        self,
        image: Path | str,
        pos: Pos | Iterable[Measure],
        size: Size | Iterable[Measure],
        scale: bool = True,
        styles: Styles = Styles(),
        parent: "Object" = None,
        components: Iterable[Component] = [],
        **_styles: Dict[str, Any]
    ):
        if not parent:
            parent = get_window()

        self.image = image
        self.scale = scale

        super().__init__(
            pos=pos,
            size=size,
            styles=styles,
            parent=parent,
            components=components,
            **_styles
        )
        try:
            self.image = pg.image.load(image)
        except:
            raise ValueError("Image not found:", image)

        if scale:
            self.image = pg.transform.scale(self.image, self.size)

    def draw(self):
        self.window.surface.blit(self.image, self.pos)

    def rotate(self, angle):
        self.image = pg.transform.rotate(self.image, angle)
        self.size = [self.image.get_width(), self.image.get_height()]

    def flip(self, x_axis: bool = True, y_axis: bool = False):
        self.image = pg.transform.flip(self.image, x_axis, y_axis)
        self.size = [self.image.get_width(), self.image.get_height()]

    def scale(self, new_size: Size):
        self.image = pg.transform.scale(self.image, new_size)
        self.size = [self.image.get_width(), self.image.get_height()]



from typing import Iterable
import pygame as pg, random, os

from .scenes import SceneManager
from .graphics import Image
from .styles.units import Measure
from .global_data import DATA, on_update
from .types import ProfilingOptions, Size, Pos
from .styles.colors import Gradient
from .styles.styles_resolver import resolve_color
from .world import World

# handlers
from .event_handler import EventHandler
from .time_handler import TimeHandler

from pstats import SortKey, Stats


pg.init()

class Window:
    """
    #### Window
    
    - Profiling: If a `ProfilingOptions` is passed to the `profiling` parameter, the profiling will be enabled.
    """
    __slots__ = (
        "size",
        "pos",
        "title",
        "icon",
        "depth",
        "vsync",
        "fullscreen",
        "resizable",
        "color",
        "surface",
        "clock",
        "show_fps",
        "fps",
        "profiling",
    )

    # check if an istance of Window is created
    is_created: bool = False

    def __new__(cls, *args, **kwargs) -> "Window":
        r"""
        #### Creates a new Window instance if it doesn't exist
        """
        if not cls.is_created:
            cls.is_created = True
            return object.__new__(cls)
        else:
            return DATA.window

    def _post_init(self):
        """
        #### Post init
        Creates global objects that need the window instance to be created beforehand.
        """
        # globalize time and event handlers
        DATA.TimeHandler = TimeHandler
        DATA.EventHandler = EventHandler
        DATA.window = self

        # set world size
        World.size = self.size

    def __init__(
        self,
        size: Size | Iterable[Measure] = Size(720, 420),
        title: str = "",
        icon: str = "",
        fps: int = 60,
        show_fps: bool = False,
        vsync: bool = False,
        depth: int = 32,
        color="black",
        fullscreen: bool = False,
        resizable: bool = False,
        profiling: ProfilingOptions = False,
    ):
        self.size = size if isinstance(size, Size) else Size(*size)
        self.pos = Pos(0, 0)
        self.title = title
        self.icon = icon
        self.vsync = vsync
        self.fullscreen = fullscreen
        self.resizable = resizable
        self.color = color
        self.fps = fps
        self.depth = depth
        self.show_fps = show_fps
        self.profiling = profiling

        self.load_icon(icon)

        # Profiling
        if self.profiling:
            self.profiling.profile.enable()

        # init window and surface
        self._init() 

        # Post init
        self._post_init()

    def __str__(self):
        return "<Window>"

    # -----------------------------------------------------------------------------
    def resize(self, size: Size):
        r"""
        #### Resizes the window
        - `size` : new size
        """
        self.size = size
        self._resolve_size(size)
        pg.display.set_mode(self.size, pg.RESIZABLE)
        return self
    
    def get_delta_time(self) -> int:
        return self.clock.get_time() / 1000

    def load_icon(self, icon: str):
        r"""
        #### Loads an icon for the window
        - `icon` :  path to the icon
        """
        self.icon = icon
        if icon == "":
            try:
                path = os.path.join(
                    os.path.dirname(__file__), "assets", "img", "icon.jpg"
                )
                self.icon = path
                pg.display.set_icon(pg.image.load(self.icon))

            except FileNotFoundError:
                pass

        return self

    def shake(self, force=5):
        r"""
        #### Shakes the window
        - `force` : force of the shake
        """
        if force <= 0:
            return

        force = int(force)
        x = random.randint(-force, force)
        y = random.randint(-force, force)
        self.surface.blit(self.surface, (x, y))
        return self

    def get_fps(self):
        r"""
        #### Returns the current window FPS
        """
        return self.clock.get_fps()

    def check_events(self):
        r"""
        #### Checks and Manage the events, should be called in the main loop
        """
        
        # run the checks in different threads to improve performance
        TimeHandler.check()
        EventHandler.check()

    def _resolve_size(self, size: Size):
        if self.fullscreen:
            self.__size = Size(size)
            self.size = pg.display.list_modes()[0]
            return

        else:
            # returns to size before fullscreen
            try:
                self.size = self.__size
                return
            except:
                pass

        if size == []:
            raise Exception("You must specify a size for the screen")

        elif len(size) == 1:
            if size[0] in ("max", "full", "100%"):
                self.size = pg.display.list_modes()[0]
            else:
                raise Exception(
                    'Screen size should "max" || "full" or list [width, height] '
                )

        elif len(size) == 2:
            if size[0] in ("max", "full", "100%"):
                self.size[0] = pg.display.list_modes()[0][0]
            elif size[1] in ("max", "full", "100%"):
                self.size[1] = pg.display.list_modes()[0][1]
            else:
                self.size = Size(size[0], size[1])

    def _init(self):
        r"""
        #### Initializes the window, is called automatically
        """

        self._resolve_size(self.size)

        if self.resizable and self.fullscreen:
            raise ValueError("You can't have resize and fullscreen at the same time")

        display_type = 0
        if self.fullscreen:
            display_type = pg.FULLSCREEN

        elif self.resizable:
            display_type = pg.RESIZABLE

        self.surface = pg.display.set_mode(
            self.size, display_type, self.depth, 0, self.vsync
        )

        pg.display.set_caption(self.title)
        if self.icon != "":
            pg.display.set_icon(pg.image.load(self.icon))

        self.clock = pg.time.Clock()


    # Property shortcuts 
    @property
    def x(self) -> Measure:
        return self.pos.x
    
    @x.setter
    def x(self, value: Measure) -> None:
        self.pos.x = value

    @property
    def y(self) -> Measure:
        return self.pos.y
    
    @y.setter
    def y(self, value: Measure) -> None:
        self.pos.y = value

    @property
    def width(self) -> Measure:
        return self.size.width
    
    @width.setter
    def width(self, value: Measure) -> None:
        self.size.width = value

    @property
    def height(self) -> Measure:
        return self.size.height
    
    @height.setter
    def height(self, value: Measure) -> None:
        self.size.height = value


    def update(self):
        r"""
        #### Updates the Window
        """

        if self.show_fps:
            pg.display.set_caption(
                f"{self.title}  FPS : " + f"{int(self.clock.get_fps())}"
            )

        pg.display.update()

        self.clock.tick(self.fps)

        # call on update events
        for func in on_update():
            func()

    def quit(self):
        r"""
        #### Quits the App  (Ends the window)
        """

        if self.profiling:
            self.profiling.profile.disable()

            # save profiling stats
            self.profiling.profile.dump_stats(self.profiling.file)

            # print profiling stats
            stats = Stats(self.profiling.file)
            stats.sort_stats(SortKey.TIME).print_stats(self.profiling.limit)

        pg.quit()
        quit()

    def fill(self, color=None, pos: list = [0, 0], size: list = [0, 0]):
        r"""
        #### Fill the Window with a `color` or `gradient`
        - `color` : color to fill the Window with, or  `Gradient` or `Image`  (Optional)
        - `pos` : position of the fill start (Optional)
        - `size` : size of the fill (Optional)
        """

        color = self.color if color == None else color

        if size == [0, 0]:
            size = self.size

        if isinstance(color, Gradient) or isinstance(color, Image):
            color.draw()

        else:
            color = resolve_color(color)
            pg.draw.rect(self.surface, color, pg.Rect(pos, size))

    def toggle_fullscreen(self):
        r"""
        #### Toggles the fullscreen mode
        """
        self.fullscreen = not self.fullscreen
        self._init()


    # Scenes
    def run_scenes(self, scene_manager: SceneManager):
        # Main loop
        while True:
            self.check_events()

            scene_manager.update()
            scene_manager.draw()
    
            self.update()



class DATA:
    on_update = {}
    window = object
    EventHandler = object
    TimeHandler = object

#shortcuts to DATA
def get_window():
    """
    #### Returns the window object
    """
    return DATA.window

def on_update(key=None,val=None):
    r"""
    #### Sets a function into `DATA.on_update` or returns all funcions if no key is passed
    """
    
    if key:
        DATA.on_update[key] = val
    else:
        return DATA.on_update.values()


from typing import Any, Dict, Iterable

from discord import Component
from ..global_data import get_window
from ..objects.object import Object
import pygame as pg
from ..styles.colors import resolve_color, Color
from ..styles.style import Styles
from ..styles.units import Measure

from ..types import Pos, Size

class Rect(Object):
    r"""
    #### Rect
    #### Parameters
    - `pos`: position of the rect `[x, y]`
    - `size`: size of the rect `[width, height]`

    #### Styles
    - `color`: color of the rect `"white" or (R, G, B)`
    - `stroke`: stroke of the rect `int`
    - `margins`: margins of the rect `[top, right, bottom, left]`
    - `border_radius`: border radius of the rect `[top-left, top-right, bottom-right, bottom-left]`
    - `components` : components to add in the rect `[Component, ..]`
    """

    def draw(self):
        pg.draw.rect(
            self.window.surface,
            self.styles.color,
            (*self.pos, *self.size),
            self.styles.stroke,
            *self.styles.border_radius
        )


class Circle(Object):
    r"""
    #### Circle
    #### Parameters
    - `pos`: position of the circle `[x, y]`
    - `radius`: radius of the circle `number`

    #### Styles
    - `color`: color of the circle `"white" or (R, G, B)`
    - `stroke`: stroke of the circle `int`
    - `margins`: margins of the circle `[top, right, bottom, left]`
    - `components` : components to add in the object `[Component, ..]`
    """

    def __init__(
        self,
        pos: Pos | Iterable[Measure],
        radius: int,
        styles: Styles = Styles(),
        parent: "Object" = None,
        components: Iterable[Component] = [],
        **_styles: Dict[str, Any]
    ):
        if not parent:
            parent = get_window()

        self.radius = radius

        super().__init__(
            pos=pos,
            size=Size(radius * 2, radius * 2),
            styles=styles,
            parent=parent,
            components=components,
            **_styles
        )

    def draw(self):
        pg.draw.circle(
            self.window.surface,
            self.styles.color,
            self.pos,
            self.radius,
            self.styles.stroke,
        )

    def _get_collision_box(self):
        center_x, center_y = self.pos
        radius = self.size[0] / 2
        return (
            (center_x - radius, center_y - radius),  # top-left
            (center_x + radius, center_y - radius),  # top-right
            (center_x - radius, center_y + radius),  # bottom-left
            (center_x + radius, center_y + radius),  # bottom-right
        )


class Ellipse(Object):
    r"""
    #### Ellipse
    #### Parameters
    - `pos`: position of the Ellipse `[x, y]`
    - `size`: size of the Ellipse `[width, height]`
    - `components` : components to add in the object `[Component, ..]`

    #### Styles
    - `color`: color of the Ellipse `"white" or (R, G, B)`
    - `stroke`: stroke of the Ellipse `int`
    - `margins`: margins of the Ellipse `[top, right, bottom, left]`
    """

    def draw(self):
        pg.draw.ellipse(
            self.window.surface,
            self.styles.color,
            (*self.pos, *self.size),
            self.styles.stroke,
        )


class Line:
    r"""
    #### Line

    #### Parameters
    - `start`: start of the line `[x, y]`
    - `end`: end of the line `[x, y]`
    - width : width of the line `int`
    - `color`: color of the line `"white" or (R, G, B)`
    - `stroke`: stroke of the line `int`

    """
    __slots__ = ("start", "end", "width", "color", "window")

    def __init__(self, start: Pos, end: Pos, width: int = 5, color: Color = "white"):
        self.start = start
        self.end = end
        self.width = width
        self.color = resolve_color(color)
        self.window = get_window()

    def draw(self):
        pg.draw.line(self.window.surface, self.color, self.start, self.end, self.width)

    def _get_collision_box(self):
        # collision is checked at line end
        return (
            self.end,
            self.end + (self.width, 0),
            self.end + (0, self.width),
            self.end + (self.width, self.width),
        )



from ast import Tuple
from typing import Callable, Iterable, List, Dict
from path import Path
import pygame
from ..styles.styles_resolver import resolve_position, resolve_size
from ..styles.units import Measure
from ezsgame.types import Pos, Size
from ..global_data import get_window
from PIL import Image, ImageSequence


pgSpriteClass = pygame.sprite.Sprite


class Sprite(pgSpriteClass):
    def __new__(
        cls,
        sprite: Path | str,
        pos: Pos | Iterable[Measure],
        size: Size | Iterable[Measure],
        scale: bool = True,
        static: bool = False,
    ):
        # if is animated
        if Path(sprite).ext == ".gif":
            raise TypeError("Use the AnimatedSprite class for animated sprites instead.")

        return object.__new__(Sprite)

    def __init__(
        self,
        sprite: Path | str,
        pos: Pos | Iterable[Measure],
        size: Size | Iterable[Measure],
        scale: bool = True,
        static: bool = False,
    ):
        pgSpriteClass.__init__(self)

        self.window = get_window()

        # resolve pos and size
        self.size = resolve_size(self, size, self.window.size, True)
        self.pos = resolve_position(self, pos, self.window, True)

        self.sprite = sprite
        self.image = pygame.image.load(sprite)

        if scale:
            self.image = pygame.transform.scale(self.image, size)

        self.rect = self.image.get_rect()
        self.rect.topleft = pos
        self.rect.size = size
        self.start_pos = pos
        self.start_size = size

        if static:
            self.draw = lambda: self.window.surface.blit(self.image, self.rect)

    def _update(self):
        self.image = pygame.transform.scale(self.image, self.size)
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.rect.size = self.size

    def draw(self):
        self._update()
        self.window.surface.blit(self.image, self.rect)


class AnimatedSprite(pgSpriteClass):
    _cached_frames: Dict[str, List[pygame.Surface]] = {} # {path: [frames]} used to cache frames/resources for animated sprites
    _cached_sprites: Dict[str, Tuple] = {} # {path: (size, frame_rate, scale, draw_method)} used to cache sprites that do the same but are in different locations
    
    def __new__(
        cls,
        sprite: Path | str,
        pos: Pos | Iterable[Measure],
        size: Size | Iterable[Measure],
        frame_rate: int = 5,
        scale: bool = True,
    ):
        if AnimatedSprite._cached_sprites.get(sprite, [None])[:-1] == (size, frame_rate, scale):
            # Return a AnimatedSpriteRef object that references the cached sprite
            return AnimatedSpriteRef.__new__(AnimatedSpriteRef, AnimatedSprite._cached_sprites[sprite][-1])
            
        return object.__new__(AnimatedSprite)

    def __init__(
        self,
        sprite: Path | str,
        pos: Pos | Iterable[Measure],
        size: Size | Iterable[Measure],
        frame_rate: int = -1, # -1 = auto
        scale: bool = True,
    ):
        pgSpriteClass.__init__(self)

        self.window = get_window()

        # resolve pos and size
        self.size = resolve_size(self, size, self.window.size, True)
        self.pos = resolve_position(self, pos, self.window, True)

        self.current_frame = 0
        self.last_update_time = 0

        if sprite in self._cached_frames:
            self.frames = self._cached_frames[sprite]

        # load frames (sprite is a gif)
        else:
            self.frames = [
                pygame.image.fromstring(frame.tobytes(), frame.size, frame.mode)
                for frame in ImageSequence.Iterator(Image.open(sprite))
            ]
            self.frames = self.frames[1:]

            self._cached_frames[sprite] = self.frames

        # set frame rate
        if frame_rate == -1:
            self.frame_rate = len(self.frames)

        # set sprite image and rect
        self.image = self.frames[self.current_frame]
        self.rect = self.image.get_rect()
        self.rect.x, self.rect.y = self.pos

        # set scale
        if scale:
            self.image = pygame.transform.scale(self.image, self.size)
            self.rect.size = self.size

        # cache sprite
        self._cached_sprites[sprite] = (size, frame_rate, scale, self.draw)

    def _update(self):
        current_time = pygame.time.get_ticks()

        # calculate time since last update
        time_since_last_update = current_time - self.last_update_time

        # update current frame if enough time has passed
        if time_since_last_update > 1000 / self.frame_rate:
            self.current_frame = (self.current_frame + 1) % len(self.frames)
            self.image = self.frames[self.current_frame]
            self.last_update_time = current_time

    def draw(self):
        self._update()
        self.window.surface.blit(self.image, self.rect)

class AnimatedSpriteRef:

    def __new__(
        cls,
        draw_method: Callable,
    ):
        return object.__new__(cls)        

    def __init__(self, draw_method: Callable):
        self.draw = draw_method

    def draw(self):
        pass


from typing import Dict, Iterable, List, Callable
import pygame as pg
from .global_data import DATA, on_update
from .objects import Object


def to_pgkey(key: str) -> int:
    '''
    #### Converts a key to a pygame key 
    Example : `a` -> `pg.K_a`
    '''
    if key.lower() == "enter":
        key = "RETURN"

    elif len(key) > 1:
        key = key.upper()

    return eval("pg.K_" + key)

class Event:
    __slot__ = "event_type", "event_name", "callback", "object", "name"
    
    def __init__(self, event_type, event_name, callback: Callable, object: Object = None, name: str = "Default", **kwargs):
        self.type = event_type
        self.event_name = event_name
        self.callback = callback
        self.object = object
        self.name = name

        for k, v in kwargs.items():
            setattr(self, k, v)

    def __contains__(self, item):
        return item in self.__dict__.keys()

    def __call__(self, **kwargs):
        try:
            self.callback(**kwargs)
        except Exception as e:
            self.callback()

class EventList(list):    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_by_type(self, event_type) -> List[Event]:
        return [event for event in self if event.type == event_type]

    def get_by_name(self, event_name) -> Event:
        lst = [event for event in self if event.event_name == event_name]
        return lst[0] if lst else None

    def replace(self, name: str, new_event: Event) -> None:
        for i, event in enumerate(self):
            if event.name == name:
                self[i] = new_event

    def remove(self, *names) -> None:
        for name in names:
            for i, event in enumerate(self):
                if event.name == name:
                    del self[i]

    def add(self, event: Event):
        # if event already exists, replace it
        if self.get_by_name(event.name):
            self.replace(event.name, event)

        # otherwise, add it
        else:
            self.append(event)


class EventHandler:
    '''
    - Manages events on the program 
    '''
    
    events = EventList()
    to_remove: List[str] = []
    to_add: List[Event] = []
    pressed_keys: pg.key.ScancodeWrapper = None

    __ezsgame_events = ("update",)

    def check():
        # gets widnow events
        events = pg.event.get()

        # log pressed keys
        EventHandler.pressed_keys = pg.key.get_pressed()

        # removes events
        for name in EventHandler.to_remove:

            for event in EventHandler.events:
                if event.name == name:
                    EventHandler.events.remove(event)
                    break

            # if is ezsgame event
            else:
                # removes event from DATA so it won't be called anymore
                if name in DATA.on_update:
                    del DATA.on_update[name]

        EventHandler.to_remove.clear()

        # adds events
        for event in EventHandler.to_add:

            # if is ezsgame event
            if event.type == "ezsgame":

                # MANAGE EZSGAME EVENTS

                # on update event
                if event.event_name == "update":
                    on_update(event.name, event.callback)

            # if is a event
            else:
                EventHandler.events.add(event)

        EventHandler.to_add.clear()

        # EVENT MANAGEMENT -------------------------------------------------------------------------------------------
        for ev in events:
            # ev : event to process

            # quit event (cannot be event listener)
            if ev.type == pg.QUIT:
                for event in EventHandler.events.get_by_type(pg.QUIT):
                    event.callback()

                DATA.window.quit()

            # Manages custom events
            for event in EventHandler.events.get_by_type("custom"):
                event_args = {
                    "key": ev.key if ev.__dict__.get("key") else None,
                    "unicode": ev.unicode if ev.__dict__.get("unicode") else None,
                    "type": ev.type,
                    "button": ev.button if ev.__dict__.get("button") else None,
                    "is_hovering": EventHandler.is_hovering(event["object"]) if "object" in event else False
                }

                event(**event_args)

            #  EVENT LOOP (managing events)
            for event in EventHandler.events.get_by_type(ev.type):

                # if is event listener (uses a object)
                is_event_listener = event.object is not None
                is_hovering = False

                if is_event_listener:
                    is_hovering = EventHandler.is_hovering(event.object)

                    # if is not hovering and event is not unhover then skip
                    if not is_hovering and not event.event_name == "unhover":
                        continue

                # function to reduce code, function decides whaterver callback should called or not
                def callback():
                    if is_event_listener:
                        # events that require be hovering
                        if is_hovering and event.object.styles.visible:
                            event.callback()

                    # if is not event listener is base event, just call callback
                    else:
                        event.callback()

                # mouse events
                if ev.type == pg.MOUSEBUTTONDOWN:
                    # mouse wheel up
                    if event.event_name == "mousewheelup" and ev.button == 4:
                        callback()
                        continue

                        # mouse wheel down
                    elif event.event_name == "mousewheeldown" and ev.button == 5:
                        callback()
                        continue
                
                    # right mouse button
                    elif event.event_name == "rightclick" and ev.button == 3:
                        callback()
                        continue
                            
                    # click, mousedown or leftclick
                    elif event.event_name in ("click", "mousedown", "leftclick") and ev.button == 1:
                        callback()
                        continue
                    

                # hover events
                elif ev.type == pg.MOUSEMOTION:
                    if event.event_name == "unhover":
                        if not is_hovering:
                            event.callback()
                            continue

                    else:
                        callback()
                        continue

                # mouse up events
                elif ev.type == pg.MOUSEBUTTONUP:
                    if event.event_name == "mouseup" or event.type == pg.MOUSEBUTTONUP:
                        event.callback()

                    else:
                        callback()

                elif "key" in event:
                    if event.key == ev.key:
                        event(key=ev.key, unicode=ev.unicode)
                        continue

                # base on key event keydown or keyapp
                elif event.type in (pg.KEYDOWN, pg.KEYUP):
                    event(key=ev.key, unicode=ev.unicode)

                # any event that matchess current window event
                else:
                    callback()

    def add_event(event: str, object: Object, callback, name: str = "Default"):
        '''
        #### Adds a event listener to a object
        - `event` : event to be added 
                - Events : `click`, `hover`, `unhover`, `unclick`.
        - `name` : name of the event 
        - `object` : object to be added to the event 
        - `callback` : function to be called when the event is triggered
        '''
        
        event, event_type = EventHandler._convert_to_pgevent(event)
        
        if name == "Default":
            name = f"{event}.{id(object)}.{len(EventHandler.events)}.{len(EventHandler.to_add)}"
            
        EventHandler.to_add.append(
            Event(event_type, event, callback, object, name))

    def remove_event(name: str):
        f'''
        #### Removes an event from the event list so it won't be called anymore
        -  `name` : name of the event to be removed 
        '''
        EventHandler.to_remove.append(name)

    def is_hovering(object: Object) -> bool:
        '''
        #### Checks if the mouse is hovering over the object
        - `object` : object to check if the mouse is hovering over it
        '''
        mouse_pos = pg.mouse.get_pos()
        box = object._get_collision_box()

        if mouse_pos[0] > box[0][0] and mouse_pos[0] < box[1][0]:
            if mouse_pos[1] > box[0][1] and mouse_pos[1] < box[2][1]:
                return True

        return False

    def on_event(event: str, callback, name: str = "Default"):
        '''
        #### Adds a `Base Event` to the event list, Calls function when event is triggered. 
        - `event`: event to be added 
                - Events : `quit`, `mousemotion`, `mousedown`, `mouseup`, `keydown`, `keyup`, `mousewheel`, `update`
        -  `callback`: function to be called when the event is triggered ``function``
        - `name`: name of event (optional)
        '''
        
        name = f"base_event.{event}.{len(EventHandler.events)}" if name == "Default" else name

        # if is ezsgame event
        if event in EventHandler.__ezsgame_events:
            EventHandler.to_add.append(
                Event("ezsgame", event, callback, None, name))
            return

        event, event_type = EventHandler._convert_to_pgevent(event)

        EventHandler.to_add.append(
            Event(event_type, event, callback, None, name))

    def on_key(type: str, keys: list, callback, name: str = "Default"):
        '''
        #### Calls function when key event is triggered.
        -  `type`: type of `Event` to be added
                        - Events : `down` (when key is down), `up` (when key released)
        - `keys`: keys to be added to the event 
        -  `callback`:  function to be called when the event is triggered 
        - `name`: name of event (optional)
        '''
        types = {
            "down": pg.KEYDOWN,
            "up": pg.KEYUP
        }

        event_type = types.get(type, None)

        if not event_type:
            raise ValueError("Invalid type: ", type)

        for key in keys:
            k = to_pgkey(key)

            name = f"{key}_{type}_{len(EventHandler.events)}" if name == "Default" else name

            EventHandler.to_add.append(
                Event(event_type, k, callback, None, name, key=k))

    def custom_event(callback, object=None, name: str = "Default"):
        '''
        #### Creates a custom event. *[Decorator]*
        - `callback` : function to be called with event parameters
        - `object` : object to check if is hovering, if you need `is_hovering` (Optional)
        - `name`: name of event (optional)
        '''

        name = f"custom_event.{name}.{len(EventHandler.events)}" if name == "Default" else name

        EventHandler.to_add.append(
            Event("custom", "custom", callback, object, name))

    def _convert_to_pgevent(event):
        if event in EventHandler.__ezsgame_events:
            return event

        event = event.lower().replace(" ", "").replace("_", "")

        evs = {
            "hover": pg.MOUSEMOTION,
            "click": pg.MOUSEBUTTONDOWN,
            "rightclick": pg.MOUSEBUTTONDOWN,
            "leftclick": pg.MOUSEBUTTONDOWN,
            "mousedown": pg.MOUSEBUTTONDOWN,
            "mouseup": pg.MOUSEBUTTONUP,
            "unhover": pg.MOUSEMOTION,
            "unclick": pg.MOUSEBUTTONUP,
            "keydown": pg.KEYDOWN,
            "keyup": pg.KEYUP,
            "mousewheelmotion": pg.MOUSEWHEEL,
            "mousemotion": pg.MOUSEMOTION,
            "quit": pg.QUIT,
            "mousebuttondown": pg.MOUSEBUTTONDOWN,
            "mousebuttonup": pg.MOUSEBUTTONDOWN,
            "mousewheelup": pg.MOUSEBUTTONDOWN,
            "mousewheeldown": pg.MOUSEBUTTONDOWN
        }

        if event not in evs:
            raise Exception("Event type not found", event)
        
        return (event, evs[event])


# event decorators ------------------------------------------------------------

def on_key(type: str, keys: Iterable | str, name: str = "Default") -> Callable:
    '''
    #### Calls the function when the key event is triggered
    - `type` : type of the event. `up` or `down`
            - Event types : `up` (when the key is released), `down` (when the key is pressed)
    - `keys` : key/keys to listen to
    - `name` : name of the event (Optional) 
    '''
    if not hasattr(keys, "__iter__") or isinstance(keys, str):
        keys = [keys]

    def wrapper(func):
        EventHandler.on_key(type, keys, func, name)
        return func

    return wrapper

def add_event(event: str, object: Object, name: str = "Default") -> Callable:
    '''
    #### Adds an event listener to an object
    - `event` : event to listen to
    - `object` : object that will be "listening"
    - `name` : name of the event (Optional)
    '''

    def wrapper(func):
        EventHandler.add_event(event, object, func, name)
        return func

    return wrapper

def on_event(event: str, name: str = "Default") -> Callable:
    '''
    #### Calls funcion when the event is triggered, (Base Event)
    - `event` : event to listen to
            - Events : `quit`, `mousemotion`, `mousedown`, `mouseup`, `keydown`, `keyup`, `mousewheel`
    - `name` : name of the event (Optional)
    '''

    if name == "Default":
        name = f"base_event.{event}.{len(EventHandler.events)}" if name == "Default" else name

    def wrapper(func):
        EventHandler.on_event(event, func, name)
        return func

    return wrapper

def custom_event(object=None, name: str = "Default") -> Callable:
    '''
    #### Adds a function as custom event
    - `object` : object to check if is hovering, if you need `is_hovering` (Optional)
    - `name` : name of the event (Optional)
    '''
    def wrapper(func):
        EventHandler.custom_event(func, object, name)
        return func

    return wrapper

def remove_event(name: str):
    '''
    #### Removes an event from the event handler
    - `name` : name of the event
    '''
    EventHandler.remove_event(name)


def is_down(key: str) -> bool:
    '''
    #### Returns if the key being is pressed
    - `key` : key to check
    '''
    return pg.key.get_pressed()[to_pgkey(key)]

def went_down(key: str) -> bool:
    '''
    #### Returns `True` if the key was pressed during this frame
    '''
    return EventHandler.pressed_keys[to_pgkey(key)]


from dataclasses import dataclass
from typing import Self, Type, TypeAlias
import math
from numpy import sort


Number: TypeAlias = float | int


class Vector2:
    """
    #### 2 Values Vector, can handle arithmetic operations
    """
    __slots__ = ("x", "y")
    
    def __init__(self, a, b = None):
        self.__call__(a, b)

    def normalize(self) -> "Vector2":
        """
        #### Normalize the vector
        """
        mag = math.sqrt(self.x**2 + self.y**2)
        if mag > 0:
            self.x /= mag
            self.y /= mag

        return self

    def dot(self, other: Type[Self]) -> Number:
        """
        #### Dot product between two vectors
        """
        return self.x * other.x + self.y * other.y
    
    def magnitude(self) -> Number:
        """
        #### Return the magnitude of the vector
        """
        return math.sqrt(self.x**2 + self.y**2)

    def to_tuple(self) -> tuple[Number, Number]:
        """
        #### Return the vector as a tuple
        """
        return (self.x, self.y)

    def __add__(a, b):
        T = a.__class__
        
        if isinstance(b, Vector2):
            return T(a.x + b.x, a.y + b.y)    

        elif isinstance(b, (int, float)):
            return T(a.x + b, a.y + b)
        
        x, y = b
        return T(a.x + x, a.y + y)

    def __iadd__(self, b):
        if isinstance(b, Vector2):
            self.x += b.x
            self.y += b.y

        elif isinstance(b, (int, float)):
            self.x += b
            self.y += b

        else:
            x, y = b
            self.x += x
            self.y += y

        return self
        
    def __sub__(a, b):
        T = a.__class__
        
        if isinstance(b, Vector2):
            return T(a.x - b.x, a.y - b.y) 

        elif isinstance(b, (int, float)):
            return T(a.x - b, a.y - b)
        
        x, y = b
        return T(a.x - x, a.y - y)
        

    def __isub__(self, b):
        if isinstance(b, Vector2):
            self.x -= b.x
            self.y -= b.y

        elif isinstance(b, (int, float)):
            self.x -= b
            self.y -= b

        else:
            x, y = b
            self.x -= x
            self.y -= y

        return self

    def __mul__(a, b):
        T = a.__class__
        
        if isinstance(b, Vector2):
            return T(a.x * b.x, a.y * b.y)  

        elif isinstance(b, (int, float)):
            return T(a.x * b, a.y * b)
        
        x, y = b
        return T(a.x * x, a.y * y)

    def __imul__(self, b):
        if isinstance(b, Vector2):
            self.x *= b.x
            self.y *= b.y

        elif isinstance(b, (int, float)):
            self.x *= b
            self.y *= b

        else:
            x, y = b
            self.x *= x
            self.y *= y

        return self

    
    def __truediv__(a, b) -> "Vector2":
        T = a.__class__
        
        if isinstance(b, Vector2):
            return T(a.x / b.x, a.y / b.y)  

        elif isinstance(b, (int, float)):
            return T(a.x / b, a.y / b)

    def __itruediv__(self, b):
        if isinstance(b, Vector2):
            self.x /= b.x
            self.y /= b.y

        elif isinstance(b, (int, float)):
            self.x /= b
            self.y /= b

        else:
            x, y = b
            self.x /= x
            self.y /= y

        return self
    
    def __floordiv__(a, b):
        T = a.__class__
        
        if isinstance(b, Vector2):
            return T(a.x // b.x, a.y // b.y)  

        elif isinstance(b, (int, float)):
            return T(a.x // b, a.y // b)
        
    def __ifloordiv__(self, b):
        if isinstance(b, Vector2):
            self.x //= b.x
            self.y //= b.y

        elif isinstance(b, (int, float)):
            self.x //= b
            self.y //= b

        else:
            x, y = b
            self.x //= x
            self.y //= y

        return self
    
    def __mod__(a, b):
        T = a.__class__
        
        if isinstance(b, Vector2):
            return T(a.x % b.x, a.y % b.y)  

        elif isinstance(b, (int, float)):
            return T(a.x % b, a.y % b)

    def __imod__(self, b):
        if isinstance(b, Vector2):
            self.x %= b.x
            self.y %= b.y

        elif isinstance(b, (int, float)):
            self.x %= b
            self.y %= b

        else:
            x, y = b
            self.x %= x
            self.y %= y

        return self
    

    def __pow__(a, b):
        T = a.__class__
        
        if isinstance(b, Vector2):
            return T(a.x ** b.x, a.y ** b.y)  

        elif isinstance(b, (int, float)):
            return T(a.x ** b, a.y ** b)
        
    def __ipow__(self, b):
        if isinstance(b, Vector2):
            self.x **= b.x
            self.y **= b.y

        elif isinstance(b, (int, float)):
            self.x **= b
            self.y **= b

        else:
            x, y = b
            self.x **= x
            self.y **= y

        return self

    def __neg__(self):
        return self.__class__(-self.x, -self.y)
    
    def __pos__(self):
        return self.__class__(+self.x, +self.y)
    
    def __abs__(self):
        return self.__class__(abs(self.x), abs(self.y))
    
    def __call__(self, a, b):
        # if a is iterable 
        if hasattr(a, "__iter__") and not isinstance(a, str):
            a, b = a

        if a != None and (b == None):
            self.x = a
            self.y = a

        else:
            self.x = a
            self.y = b

    def __str__(self):
        return f"<Vector2 : {self.x}, {self.y}>"

    def __repr__(self):
        return f"Vector2({self.x}, {self.y})"

    def __iter__(self):
        return (self.x, self.y).__iter__() 

    def __getitem__(self, index):
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        else:
            raise IndexError

    def __setitem__(self, index, value):
        if index == 0:
            self.x = value
        elif index == 1:
            self.y = value
        else:
            raise IndexError

    def __len__(self):
        return 2

    def copy(self) -> Type[Self]:
        return self.__class__(self.x, self.y)

    def ref(self):
        return self

    def __eq__(self, other):
        if isinstance(other, Vector2):
            return self.x == other.x and self.y == other.y
        else:
            return False

    def __ne__(self, other):
        return not self == other
    
    def set(self, a, b):
        self.x = a
        self.y = b
    
class Size (Vector2):
    r"""
    #### Size
    #### Parameters
    - `width`: width  `int` or `[width, height]`
    - `height`: height `int` or `[width, height]`
    """

    def __init__(self, width: Number, height: Number = None):
        super().__init__(width, height)

    @property
    def width(self) -> Number:
        return self.x

    @width.setter
    def width(self, value):
        self.x = value

    @property
    def height(self) -> Number:
        return self.y

    @height.setter
    def height(self, value):
        self.y = value

    def __str__(self):
        return f"(width: {self.width}, height: {self.height})"
    
    def __repr__(self):
        return f"Size({self.width}, {self.height})"

class Pos (Vector2):
    r"""
    #### Position
    #### Parameters
    - `x`: x position `number` or `[x, y]`
    - `y`: y position `number`
    """

    def __init__(self, x: Number, y: Number = None):
        super().__init__(x, y)

    def __str__(self):
        return f"(x: {self.x}, y: {self.y})"

    def __repr__(self):
        return f"Pos({self.x}, {self.y})"

# Profiling options
import cProfile
from pstats import SortKey

@dataclass
class ProfilingOptions:
    profile: cProfile.Profile = cProfile.Profile()
    sort : SortKey = SortKey.CUMULATIVE
    limit : int = 10
    file : str = "profile.prof"


from typing import Any, Callable, Dict, Iterable, Optional, Self, Set, Type
import pygame as pg

from ezsgame.styles.units import Measure

from ..styles.style import Styles

from ..components import ComponentGroup, Component
from ..global_data import get_window
from ..styles.styles_resolver import resolve_position, resolve_size
from ..funcs import center_at
from ..reactivity import Reactive

from ..types import Number, Pos, Size
from ..world import World

class Object:
    r"""
    Object is a base class for all objects in the game.
    should not be instantiated.
    """

    __slots__ = ("pos", "size", "window", "components", "behavior", "__on_draw", "parent", "children", "styles")

    def __init__(
        self, 
        pos: Pos | Iterable[Measure],
        size: Size | Iterable[Measure],
        styles: Optional[Styles] = None,
        parent: Optional["Object"] = None,
        components: Iterable[Component] = [],
        **_styles: Dict[str, Any]
    ):
        """
        Base class for most object, handlers object default and required behavior
        """
        self.window = get_window()
        self.children: Set[Object] = set()
        
        if parent:
            self.parent = parent
            
            if parent != self.window:
                self.parent.add_child(self)
            
        else:
            self.parent = self.window
        
        
        self.styles = styles or Styles(**_styles)
        # resolve styles 
        self.styles.resolve(self.parent.size)
        
        self.size = resolve_size(self, size, self.parent.size)
        self.pos = resolve_position(self, pos, self.parent)

        # Adapt position to world
        self.pos += World.pos


        # defualt behavior - needs it own type and rework
        self.behavior = {"pos": "dynamic"}

        self.__on_draw: Dict[str, Callable] = {}

        self.components = ComponentGroup(self, *components)

        # Calls __after_draw() after  draw() to call after-draw methods
        def _draw_manager(draw_func):
            def wrapper():
                if self.styles.visible:
                    draw_func() 
                                       
                self.__after_draw()
                
            return wrapper
        try:
            self.draw = _draw_manager(self.draw)
        except:
            pass

    def add_child(self, child: Type[Self]) -> None:
        self.children.add(child)

    def _update(self, updated_property_name: str) -> None:
        """
        This method is called when a Reactive property (binded to this object) is updated.
        """
        value: Reactive = getattr(self, updated_property_name)
        value = value.get()
        
        if self.parent != self.window:
            self.parent._child_update(self) 
            
            
    def _child_update(self, child: Self) -> None:
        """
        Called when a child is updated.
        """
        pass        

    def on_draw(self, func, name: str = "Default", pass_object=False):
        r"""
        #### Adds a function to be called after the object is drawn
        #### Parameters
        - `func`: function to be called
        - `name`: name of the function (default: `function name`)
        - `pass_object`: if True, the function will be called with the object as the first parameter

        """
        name = name if name else func.__name__
        self.__on_draw[name] = lambda: func(self) if pass_object else func

    def remove_on_draw(self, name: str):
        r"""
        #### Removes a function from being called after the object is drawn
        #### Parameters
        - `name`: name of the function
        """
        if name in self.__on_draw:
            del self.__on_draw[name]

    def _get_collision_box(self):
        x, y = self.pos
        w, h = self.size
        return [(x, y), (x + w, y), (x, y + h), (x + w, y + h)]

    def __str__(self):
        return f"<Object: {self.__class__.__name__}, ID: {id(self)}>"

    def __after_draw(self):
        r'''
        manages Object draw method and calls `on_draw` functions
        '''
        for func in self.__on_draw.values():
            func()
               
    def center_at(self, object: Type[Self] = None) -> Self:
        r'''
        #### Centers the object at another object
        #### Parameters
        - `object`: object to center at
        '''
        object = object or self.parent
        center_at(self, object)
        return self
            
    
    @property
    def center(self) -> Pos:
        r'''
        #### Returns the center of the object
        '''
        return self.pos + self.size / 2
    
    @center.setter
    def center(self, value: Pos) -> None:
        r'''
        #### Sets the center of the object
        #### Parameters
        - `value`: center position
        '''
        self.pos = value - self.size / 2

    @property
    def width(self) -> Number:
        return self.x

    @width.setter
    def width(self, value):
        self.x = value

    @property
    def height(self) -> Number:
        return self.y

    @height.setter
    def height(self, value):
        self.y = value

    @property
    def x(self) -> Number:
        return self.pos[0]

    @x.setter
    def x(self, value):
        self.pos[0] = value

    @property
    def y(self) -> Number:
        return self.pos[1]

    @y.setter
    def y(self, value):
        self.pos[1] = value    


#  Main
from .window import *
from .objects import *
from .event_handler import *
from .time_handler import *

from .global_data import get_window

# Resources
from .sounds import *
from .funcs import *
from .mouse import *
from .types import *
from .utils import *
from .graphics import *

from .styles.colors import *
from .styles.style import *

# Extras
from . import camera
from .scenes import *


# Furures
from .components import *
from .futures.controller import Controller
from .futures.hot_reload import *


from typing import Optional
import pygame as pg

class Mixer:
    def __init__(self):
        if pg.mixer.get_init() is None:
            pg.mixer.init()
    
        self.sounds = []
        
    def __str__(self):
        return f"<Object: Mixer, ID : {self._id}>"

    def __repr__(self):
        return f"<Object: Mixer, ID : {self._id}>"
        
    def load(self, filename):
        self.sounds.append(Sound(filename))
        
    def _load_sound(self, sound_object):
        self.sounds.append(sound_object)

    def play(self):
        for sound in self.sounds:
            sound.play()

    def stop(self):
        for sound in self.sounds:
            sound.stop()
            
    def remove(self, sound):
        if sound in self.sounds:
            self.sounds.remove(sound)
        else:
            raise ValueError("Sound not in list")
     
    def clear(self):
        for sound in self.sounds:
            del sound
  
class Sound:
    defualt_mixer = Mixer()
    
    def __init__(self, file, mixer: Optional[Mixer]):
        try:
            self.sound = pg.mixer.Sound(file)
        except Exception as e:
            raise Exception(f"Could not load sound file <{file}>. \n Error: {e}")
        
        self.sound.set_volume(0.5)
        self.volume = 0.5
        
        self.mixer = mixer or Sound.defualt_mixer
        
        self.mixer._load_sound(self)
        self.length = self.sound.get_length()
        
    @property
    def volume(self):
        return self.sound.get_volume()

    @volume.setter
    def volume(self, volume):
        self.sound.set_volume(volume)

    def play(self):
        self.sound.play()

    def stop(self):
        self.sound.stop()
        
    def set_volume(self, volume):
        self.sound.set_volume(volume)
    
    def fadeout(self, time):
        self.sound.fadeout(time)
        
    def fadein(self, time):
        self.sound.fadein(time)
        
    def __str__(self):
        return f"<Object: Sound>"

    def __repr__(self):
        return f"<Object: Sound>"
    
    def __del__(self):
        if self in self.mixer.sounds:
            self.mixer.remove(self)
        del self.sound
        
    def raw(self):
        return self.sound.get_raw()


from typing import Any, Tuple, Union
from ..funcs import center_at

from ..objects.object import Object
from ..types import Pos

class Group:
    r"""
    #### Group
    
	#### Parameters
    - `parent` : parent of the group, objects position in the group will be relative to this object `Object`
        - if parent is a string, it will be searched in the group items
        
    - [*args] `objects` : objects to add in the group `Object, ..`
    - [**kwargs] `named_objects` : named objects to add in the group `name = Object, ..`

    #### Notes
    - named objects can be accessed by: `group["name"]`
    - no named objects can be accessed by: `group[id(object)]`

    """
    
    def __init__(self, *objects, **named_objects):
        self.__objects = {}
        
        parent = named_objects.get("parent", None)
        if parent:
            del named_objects["parent"]
        
        self.add(*objects, **named_objects)
        
        if parent:
            if isinstance(parent, str):
                parent = self.__objects[parent]
                
        else:
            parent = None

        self.__parent = parent
        self.__last_parent_pos = Pos(0,0)
        
    def add(self, *objects, **named_objects):
        for obj in objects:
            self.__objects[id(obj)] = obj
        
        self.__objects.update(named_objects)

    def add_as(self, key: Union[int, float, bool, str, Tuple], object: Object):
        """
        Adds a object to the group with a custom key.
        Make sure the key is a valid dictionary key.        
        """
        self.__objects[key] = object

    def get(self, name: Union[int, float, bool, str, Tuple], default: Any = None) -> Object | None:
        """
        Gets a object from the group by it's name if it exists, otherwise `default` will be returned.
        """
        return self.__objects.get(name, default)
    
    def remove(self, name: str):
        self.__objects.pop(name)

    def align_objects(self, auto_size = True):
        # aligns objects in the group
        if self.__parent.pos != self.__last_parent_pos:
                
            margin = self.__parent.size[1]*0.1
                
            current_y = self.__parent.pos[1] + margin
            
            for obj in self.__objects.values():
                if obj != self.__parent:
                
                    center_at(obj, self.__parent)
                    obj.pos[1] = current_y
                
                    current_y += obj.size[1] + margin
                    
                    if auto_size:
                        # if overflow, resize parent
                        
                        # x axis
                        if obj.pos[0] + obj.size[0] > self.__parent.pos[0] + self.__parent.size[0]:
                            self.__parent.size[0] = obj.pos[0] + obj.size[0] + margin * 1.5
                            center_at(obj, self.__parent)
                            
                        # y axis
                        if obj.pos[1] + obj.size[1] > self.__parent.pos[1] + self.__parent.size[1]:
                            self.__parent.size[1] = obj.pos[1] + obj.size[1] + margin * 1.5

            self.__last_parent_pos = self.__parent.pos

    def draw(self):
        # align objects to paren    
        if self.__parent:
            self.align_objects()
        
        for obj in self.values():
            obj.draw()

    def map(self, func):
        for obj in self.values():
            func(obj)

    def __str__(self):
        t = ", ".join([str(i) for i in self.objects])
        return f"<Group : {t} >"

    def __repr__(self):
        return self.__str__()

    def filter(self, func) -> 'Group':
        d = {k: v for k, v in self.items() if func(v)}
        return d


    # getters        
    def __getitem__(self, key):
        if isinstance(key, int):
            return self.__objects[list(self.__objects.keys())[key]]
        return self.__objects[key]

    def __getattr__(self, name):
        return self.__objects.get(name, None) or self.__dict__[name]
    
    def values(self, no_parent=False):
        if no_parent and self.__parent:
            return [v for k, v in self.__objects.items() if v != self.__parent]
        
        return self.__objects.values()
    
    def items(self, no_parent=False):
        if no_parent and self.__parent:
            return {k: v for k, v in self.__objects.items() if v != self.__parent}
        
        return self.__objects.items()

    def keys(self, no_parent=False):
        if no_parent and self.__parent:
            return [k for k, v in self.__objects.items() if v != self.__parent]
    
        return self.__objects.keys()
    

    def __del__(self):
        for obj in self.__objects.values():
            del obj
        

    # delete item
    def __delitem__(self, key):
        self.__objects.pop(key)
        self.__dict__.pop(key)
        
    def __delattr__(self, key):
        self.__objects.pop(key)
        self.__dict__.pop(key)
    
    # contains
    def __contains__(self, key):
        return key in self.__objects
    
    # iterables
    def __iter__(self):
        return iter(self.values())
    
    def __next__(self):
        return next(self.values())

    def __len__(self):
        return len(self.__objects)


from typing import Any, Dict, Iterable, Set
import pygame as pg
from ..components import Component
from ..global_data import get_window
from ..objects.object import Object
from ..reactivity import Reactive
from ..styles.style import Styles
from ..styles.units import Measure
from ..types import Pos, Size
from ..fonts import Fonts, FontFamily


pg.font.init()


class Text(Object):
    r"""
    #### Text
    #### Parameters
    - `text`: text to be displayed
    - `pos`: position of the text `[x, y]`
    - `font_size`: font size of the text
    - `font` : font of the text `"Arial or "path/to/font.ttf"` or `ezsgame font` (default: `OpenSans`)
    - `bold` : if True, the text will be bold `bool`
    - `italic` : if True, the text will be italic `bool`

    #### Styles
    - `color`: color of the text `"white" or (R, G, B)`
    - `margins`: margins of the text `[top, right, bottom, left]`
    - `components` : components to add in the object `[Component, ..]`
    """

    __slots__ = (
        "font",
        "font_size",
        "text",
        "bold",
        "italic",
        "text_obj",
        "styles",
        "children",
    )

    def __init__(
        self,
        text: str,
        pos: Pos | Iterable[Measure],
        font_size: int,
        styles: Styles = Styles(),
        font: FontFamily | str = Fonts.OpenSans,
        parent: Object = None,
        components: Iterable[Component] = [],
        italic: bool = False,
        bold: bool = False,
        **_styles: Dict[str, Any],
    ):
        if not parent:
            parent = get_window()

        self.children: Set[Object] = set()

        self.font = font
        self.font_size = Reactive(font_size)._mount(self, "font_size")
        self.text = Reactive(text)._mount(self, "text")

        self.bold = bold
        self.italic = italic

        # need before supert init because load_font uses it
        self.styles = styles
        self.styles.resolve(parent.size)

        self.text_obj = self.load_font()

        super().__init__(
            pos=pos,
            size=Size(self.text_obj.get_width(), self.text_obj.get_height()),
            components=components,
            parent=parent,
            styles=styles,
            **_styles,
        )

    def load_font(self):
        # is font is a ezsgame font
        if isinstance(self.font, FontFamily):
            font = self.font.get_font(self.font_size.get())

        # if font is a path | str
        elif isinstance(self.font, str):
            # if font in system fonts
            if font in pg.font.get_fonts():
                font = pg.font.SysFont(
                    font, self.font_size.get(), self.bold, self.italic
                )

            # if font is a path
            elif font.endswith(".ttf"):
                try:
                    font = pg.font.Font(font, self.font_size.get())
                except Exception as e:
                    raise ValueError(f"Error loading font: {e}")

            else:
                raise ValueError("Invalid font name or path: " + self.font)

        else:
            raise ValueError("Invalid font: " + self.font)

        return font.render(self.text.get(), True, self.styles.color)

    def _update(self, updated_property_name: str):
        if updated_property_name == "text":
            self.text_obj = self.load_font()
            self.size = Size(self.text_obj.get_width(), self.text_obj.get_height())

        super()._update(updated_property_name)

    def draw(self):
        self.window.surface.blit(self.text_obj, self.pos)



from typing import Tuple
import pygame as pg

from ezsgame.types import Pos
from ezsgame.world import World


def get_mouse_pos() -> Pos:
    r'''
    #### Returns the mouse position
    Note: position is already adapted to the world position
    '''
    return Pos(pg.mouse.get_pos()) + World.pos


from typing import Any, Callable, List, Self, Set


class Reactive:
    def __init__(self, value: Any) -> None:

        if isinstance(value, Reactive):
            self = value
            
        else:
            self._value = value
            
            self._update_method: Callable[[str], None] = None
            self._property_name: str = None
            
    def _mount(self, object, property_name: str, update_method_name: str = "_update") -> Self:
        
        self._update_method = getattr(object, update_method_name, None)
        self._property_name = property_name
        
        if not self._update_method:
            raise ValueError(f"Error while trying to mount Reactive value -> Cannot find {update_method_name} method at {object}")
        
        return self
        
    def set(self, new_value: Any) -> None:
        self._value = new_value
        self._update_method(self._property_name)

    def get(self) -> Any:
        return self._value
    

    # Arithmetic magic method that change the value
    def __iadd__(self, other):
        self._value += other
        self.set(self._value)
        return self

    def __isub__(self, other):
        self._value -= other
        self.set(self._value)
        return self

    def __imul__(self, other):
        self._value *= other
        self.set(self._value)
        return self

    def __itruediv__(self, other):
        self._value /= other
        self.set(self._value)
        return self


    # Arithmetic magic method
    def __add__(self, other):
        return self._value + other
    
    def __sub__(self, other):
        return self._value - other

    def __mult__(self, other):
        return self._value * other
    
    def __div__(self, other):
        return self._value / other
    
    def __truediv__(self, other):
        return self._value // other
    
    def __pow__(self, other):
        return self._value ** other

    # Comparing magic methods
    
    def __eq__(self, other):
        return self._value == other

    def __ne__(self, other):
        return self._value != other

    def __lt__(self, other):
        return self._value < other
    
    def __le__(self, other):
        return self._value <= other

    def __gt__(self, other):
        return self._value > other

    def __ge__(self, other):
        return self._value >= other

    def __neg__(self):
        return -self._value

    def __pos__(self):
        return +self._value

    def __abs__(self):
        return abs(self._value)

    def __invert__(self):
        return ~self._value

    def __and__(self, other):
        return self._value & other

    def __or__(self, other):
        return self._value | other

    def __xor__(self, other):
        return self._value ^ other

    def __lshift__(self, other):
        return self._value << other

    def __rshift__(self, other):
        return self._value >> other



import pygame
from abc import ABC, abstractmethod
from .global_data import get_window

Vector2 = pygame.math.Vector2

class Camera:
    def __init__(self, object, scrollmethod=None):
        self.object = object
        self.offset = Vector2(0, 0)
        self.width, self.height = get_window().size
        self.CONST = Vector2(-self.width / 2 + object.size[0] / 2, -self.object.size[1] + 20)

        if scrollmethod:
            self.setmethod(scrollmethod)

    def setmethod(self, method):
        try:
            self.method = method(self)
        except:
            self.method = method

    def scroll(self):
        self.method.scroll()

class CamScroll(ABC):
    def __call__(self, camera):
        self.camera = camera
        self.object = camera.object
    
    def __init__(self, camera):
        self.__call__(camera)

    @abstractmethod
    def scroll(self):
        pass

class Follow(CamScroll):
    def __init__(self, camera=None):
        CamScroll.__init__(self, camera)

    def scroll(self):
        self.camera.offset.x += (self.object.pos.x - self.camera.offset.x + self.camera.CONST.x)
        self.camera.offset.y += (self.object.pos.y - self.camera.offset.y + self.camera.CONST.y)
        self.camera.offset.x, self.camera.offset.y = int(self.camera.offset.x), int(self.camera.offset.y)

class Border(CamScroll):
    def __init__(self, camera=None, borders=[]):
        self.borders = borders
        
        if not borders:
            self.borders = [self.object.pos.x, self.object.pos.x + self.object.size.width, self.object.pos.y, self.object.pos.y + self.object.size.height]
        
        CamScroll.__init__(self, camera)
        
    def scroll(self):
        self.camera.offset.x += (self.object.pos.x - self.camera.offset.x + self.camera.CONST.x)
        self.camera.offset.y += (self.object.pos.y - self.camera.offset.y + self.camera.CONST.y)
        self.camera.offset.x, self.camera.offset.y = int(self.camera.offset.x), int(self.camera.offset.y)
        
        #border right 
        self.camera.offset.x = max(self.borders[0], self.camera.offset.x)
        
        #   border left
        self.camera.offset.x = min(self.camera.offset.x, self.borders[1] - self.camera.width)

class Auto(CamScroll):
    def __init__(self,camera=None):
        CamScroll.__init__(self,camera)

    def scroll(self):
        self.camera.offset.x += 1




from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Optional, Self, Set, Tuple, Type, Union
import pygame as pg

from ezsgame.styles.units import Measure

from .styles.style import Styles

from .futures.components import ComponentGroup, Component
from .global_data import get_id, get_window
from .styles.styles_resolver import resolve_color, Color, resolve_position, resolve_size
from .funcs import center_at
from .fonts import Fonts, FontFamily
from .reactivity import Reactive

from .types import Number, Pos, Size

# Initialize vars ---------------------------------------------------------
pg.font.init()


# Objects -----------------------------------------------------------------------
class Object:
    r"""
    Object is a base class for all objects in the game.
    should not be instantiated.
    """

    __slots__ = ("pos", "size", "window", "id", "components", "behavior", "__on_draw", "parent", "children", "styles")

    def __init__(
        self, 
        pos: Pos | Iterable[Measure],
        size: Size | Iterable[Measure],
        styles: Optional[Styles] = None,
        parent: Optional["Object"] = None,
        components: Iterable[Component] = [],
        **_styles: Dict[str, Any]
    ):
        """
        Base class for most object, handlers object default and required behavior
        """
        self.window = get_window()
        self.id = get_id()
        self.children: Set[Object] = set()
        
        if parent:
            self.parent = parent
            
            if parent != self.window:
                self.parent.add_child(self)
            
        else:
            self.parent = self.window
        
        
        self.styles = styles or Styles(**_styles)
        # resolve styles 
        self.styles.resolve(self.parent.size)
        
        self.size = resolve_size(self, size, self.parent.size)
        self.pos = resolve_position(self, pos, self.parent)


        # defualt behavior - needs it own type and rework
        self.behavior = {"pos": "dynamic"}

        self.__on_draw: Dict[str, Callable] = {}

        self.components = ComponentGroup(self, *components)

        # Calls __after_draw() after  draw() to call after-draw methods
        def _draw_manager(draw_func):
            def wrapper():
                if self.styles.visible:
                    draw_func() 
                                       
                self.__after_draw()
                
            return wrapper
        try:
            self.draw = _draw_manager(self.draw)
        except:
            pass

    def add_child(self, child: Type[Self]) -> None:
        self.children.add(child)

    def _update(self, updated_property_name: str) -> None:
        """
        This method is called when a Reactive property (binded to this object) is updated.
        """
        value: Reactive = getattr(self, updated_property_name)
        value = value.get()
        
        if self.parent != self.window:
            self.parent._child_update(self) 
            
            
    def _child_update(self, child: Self) -> None:
        """
        Called when a child is updated.
        """
        return        

    def on_draw(self, func, name: str = "Default", pass_object=False):
        r"""
        #### Adds a function to be called after the object is drawn
        #### Parameters
        - `func`: function to be called
        - `name`: name of the function (default: `function name`)
        - `pass_object`: if True, the function will be called with the object as the first parameter

        """
        name = name if name else func.__name__
        self.__on_draw[name] = lambda: func(self) if pass_object else func

    def remove_on_draw(self, name: str):
        r"""
        #### Removes a function from being called after the object is drawn
        #### Parameters
        - `name`: name of the function
        """
        if name in self.__on_draw:
            del self.__on_draw[name]

    def _get_collision_box(self):
        x, y = self.pos
        w, h = self.size
        return [(x, y), (x + w, y), (x, y + h), (x + w, y + h)]

    def __str__(self):
        return f"<Object: {self.__class__.__name__}, ID: {self.id}>"

    def __after_draw(self):
        r'''
        manages Object draw method and calls `on_draw` functions
        '''
        for func in self.__on_draw.values():
            func()
               
    def center_at(self, object: Type[Self] = None) -> Self:
        r'''
        #### Centers the object at another object
        #### Parameters
        - `object`: object to center at
        '''
        object = object or self.parent
        center_at(self, object)
        return self
            
    
    @property
    def center(self) -> Pos:
        r'''
        #### Returns the center of the object
        '''
        return self.pos + self.size / 2
    
    @center.setter
    def center(self, value: Pos) -> None:
        r'''
        #### Sets the center of the object
        #### Parameters
        - `value`: center position
        '''
        self.pos = value - self.size / 2

    @property
    def width(self) -> Number:
        return self.x

    @width.setter
    def width(self, value):
        self.x = value

    @property
    def height(self) -> Number:
        return self.y

    @height.setter
    def height(self, value):
        self.y = value

    @property
    def x(self) -> Number:
        return self.pos[0]

    @x.setter
    def x(self, value):
        self.pos[0] = value

    @property
    def y(self) -> Number:
        return self.pos[1]

    @y.setter
    def y(self, value):
        self.pos[1] = value    


class Rect(Object):
    r'''
    #### Rect
    #### Parameters
    - `pos`: position of the rect `[x, y]`
    - `size`: size of the rect `[width, height]`

    #### Styles
    - `color`: color of the rect `"white" or (R, G, B)`
    - `stroke`: stroke of the rect `int`
    - `margins`: margins of the rect `[top, right, bottom, left]`
    - `border_radius`: border radius of the rect `[top-left, top-right, bottom-right, bottom-left]`
    - `components` : components to add in the rect `[Component, ..]`
    '''         

    def draw(self):
        pg.draw.rect(
            self.window.surface, 
            self.styles.color, 
            (*self.pos, *self.size), 
            self.styles.stroke, 
            *self.styles.border_radius
        )

class Text(Object):
    r'''
    #### Text
    #### Parameters
    - `text`: text to be displayed
    - `pos`: position of the text `[x, y]`
    - `font_size`: font size of the text
    - `font` : font of the text `"Arial or "path/to/font.ttf"` or `ezsgame font` (default: `OpenSans`)
    - `bold` : if True, the text will be bold `bool`    
    - `italic` : if True, the text will be italic `bool`

    #### Styles
    - `color`: color of the text `"white" or (R, G, B)`
    - `margins`: margins of the text `[top, right, bottom, left]`
    - `components` : components to add in the object `[Component, ..]`
    '''

    __slots__ = ("font", "font_size", "text", "bold", "italic", "text_obj", "styles", "children")

    def __init__(
        self, 
        text: str, 
        pos: Pos | Iterable[Measure], 
        font_size: int, 
        styles: Styles = Styles(),
        font: FontFamily | str = Fonts.OpenSans,
        parent: Object = None,
        components: Iterable[Component] = [],
        italic: bool = False,
        bold: bool = False,
        **_styles: Dict[str, Any]
    ):
        
        if not parent:
            parent = get_window()
        
        self.children: Set[Object] = set()
        
        self.font = font
        self.font_size = Reactive(font_size)._mount(self, "font_size")
        self.text = Reactive(text)._mount(self, "text")
        
        self.bold = bold
        self.italic = italic
        
        # need before supert init because load_font uses it
        self.styles = styles 
        self.styles.resolve(parent.size)
        
        self.text_obj = self.load_font()

        super().__init__(
            pos=pos, 
            size=Size(self.text_obj.get_width(), self.text_obj.get_height()), 
            components=components, 
            parent=parent, 
            styles=styles,
            **_styles
        )
        
    def load_font(self):
        # is font is a ezsgame font
        if isinstance(self.font, FontFamily):
            font = self.font.get_font(self.font_size.get())
        
        # if font is a path | str
        elif isinstance(self.font, str):
            
            # if font in system fonts
            if font in pg.font.get_fonts():
                font = pg.font.SysFont(font, self.font_size.get(), self.bold, self.italic)

            # if font is a path
            elif font.endswith(".ttf"):
                try:
                    font = pg.font.Font(font, self.font_size.get())
                except Exception as e:
                    raise ValueError(f"Error loading font: {e}")
                
            else:
                raise ValueError("Invalid font name or path: " + self.font) 
            
        else:
            raise ValueError("Invalid font: " + self.font)
        
        return font.render(self.text.get(), True, self.styles.color)

    def _update(self, updated_property_name: str):      
          
        if updated_property_name == "text":
            self.text_obj = self.load_font()
            self.size = Size(self.text_obj.get_width(), self.text_obj.get_height())
            
        super()._update(updated_property_name)


    def draw(self):
        self.window.surface.blit(self.text_obj, self.pos)

class Image(Object):
    r'''
    #### Image
    #### Parameters
    - `pos`: position of the image `[x, y]`
    - `size` : size of the image `[width, height]`
    - `image` : path to image file `str`

    #### Optional Arguments
    - `scale` : scale the image to the size `bool`
    - `components` : components to add in the object `[Component, ..]`
    - `styles` : Styles
    '''

    def __init__(
        self, 
        image: Path | str,
        pos: Pos | Iterable[Measure],  
        size: Size | Iterable[Measure],
        scale: bool = True,
        styles: Styles = Styles(),
        parent: "Object" = None,
        components: Iterable[Component] = [],
        **_styles: Dict[str, Any]
    ):
        if not parent:
            parent = get_window()
        
        self.image = image
        self.scale = scale
        
        super().__init__(pos=pos, size=size, styles=styles, parent=parent, components=components, **_styles)
        try:
            self.image = pg.image.load(image)
        except:
            raise ValueError("Image not found:", image)

        if scale:
            self.image = pg.transform.scale(self.image, self.size)

    def draw(self):
        self.window.surface.blit(self.image, self.pos)

    def rotate(self, angle):
        self.image = pg.transform.rotate(self.image, angle)
        self.size = [self.image.get_width(), self.image.get_height()]

    def flip(self, x_axis: bool = True, y_axis: bool = False):
        self.image = pg.transform.flip(self.image, x_axis, y_axis)
        self.size = [self.image.get_width(), self.image.get_height()]

    def scale(self, new_size: Size):
        self.image = pg.transform.scale(self.image, new_size)
        self.size = [self.image.get_width(), self.image.get_height()]

class Circle(Object):
    r'''
    #### Circle
    #### Parameters
    - `pos`: position of the circle `[x, y]`
    - `radius`: radius of the circle `number`

    #### Styles
    - `color`: color of the circle `"white" or (R, G, B)`
    - `stroke`: stroke of the circle `int`
    - `margins`: margins of the circle `[top, right, bottom, left]`
    - `components` : components to add in the object `[Component, ..]`
    '''

    def __init__(
        self, 
        pos: Pos | Iterable[Measure],
        radius: int,
        styles: Styles = Styles(),
        parent: "Object" = None,
        components: Iterable[Component] = [],
        **_styles: Dict[str, Any]
        
    ):
        if not parent:
            parent = get_window()
        
        self.radius = radius

        super().__init__(pos=pos, size=Size(radius*2, radius*2), styles=styles, parent=parent, components=components, **_styles)

    def draw(self):
        pg.draw.circle(
            self.window.surface,
            self.styles.color,
            self.pos, 
            self.radius, 
            self.styles.stroke
        )

    def _get_collision_box(self):
        center_x, center_y = self.pos
        radius = self.size[0]/2
        return (
            (center_x - radius, center_y - radius), # top-left
            (center_x + radius, center_y - radius), # top-right
            (center_x - radius, center_y + radius), # bottom-left
            (center_x + radius, center_y + radius)  # bottom-right
        )

class Ellipse(Object):
    r'''
        #### Ellipse
        #### Parameters
        - `pos`: position of the Ellipse `[x, y]`
        - `size`: size of the Ellipse `[width, height]`
        - `components` : components to add in the object `[Component, ..]`

        #### Styles
        - `color`: color of the Ellipse `"white" or (R, G, B)`
        - `stroke`: stroke of the Ellipse `int`
        - `margins`: margins of the Ellipse `[top, right, bottom, left]`
        '''

    def draw(self):
        pg.draw.ellipse(
            self.window.surface,
            self.styles.color, 
            (*self.pos, *self.size), 
            self.styles.stroke
        )

class Group:
    r"""
    #### Group
    
	#### Parameters
    - `parent` : parent of the group, objects position in the group will be relative to this object `Object`
        - if parent is a string, it will be searched in the group items
        
    - [*args] `objects` : objects to add in the group `Object, ..`
    - [**kwargs] `named_objects` : named objects to add in the group `name = Object, ..`

    #### Notes
    - named objects can be accessed by: `group["name"]`
    - no named objects can be accessed by: `group[object.id]`

    """
    
    def __init__(self, *objects, **named_objects):
        self.__objects = {}
        
        parent = named_objects.get("parent", None)
        if parent:
            del named_objects["parent"]
        
        self.add(*objects, **named_objects)
        
        if parent:
            if isinstance(parent, str):
                parent = self.__objects[parent]
                
        else:
            parent = None

        self.__parent = parent
        self.__last_parent_pos = Pos(0,0)
        
    def add(self, *objects, **named_objects):
        for obj in objects:
            self.__objects[obj.id] = obj
        
        self.__objects.update(named_objects)

    def add_as(self, key: Union[int, float, bool, str, Tuple], object: Object):
        """
        Adds a object to the group with a custom key.
        Make sure the key is a valid dictionary key.        
        """
        self.__objects[key] = object

    def get(self, name: Union[int, float, bool, str, Tuple], default: Any = None) -> Object | None:
        """
        Gets a object from the group by it's name if it exists, otherwise `default` will be returned.
        """
        return self.__objects.get(name, default)
    
    def remove(self, name: str):
        self.__objects.pop(name)

    def align_objects(self, auto_size = True):
        # aligns objects in the group
        if self.__parent.pos != self.__last_parent_pos:
                
            margin = self.__parent.size[1]*0.1
                
            current_y = self.__parent.pos[1] + margin
            
            for obj in self.__objects.values():
                if obj != self.__parent:
                
                    center_at(obj, self.__parent)
                    obj.pos[1] = current_y
                
                    current_y += obj.size[1] + margin
                    
                    if auto_size:
                        # if overflow, resize parent
                        
                        # x axis
                        if obj.pos[0] + obj.size[0] > self.__parent.pos[0] + self.__parent.size[0]:
                            self.__parent.size[0] = obj.pos[0] + obj.size[0] + margin * 1.5
                            center_at(obj, self.__parent)
                            
                        # y axis
                        if obj.pos[1] + obj.size[1] > self.__parent.pos[1] + self.__parent.size[1]:
                            self.__parent.size[1] = obj.pos[1] + obj.size[1] + margin * 1.5

            self.__last_parent_pos = self.__parent.pos

    def draw(self):
        # align objects to paren    
        if self.__parent:
            self.align_objects()
        
        for obj in self.values():
            obj.draw()

    def map(self, func):
        for obj in self.values():
            func(obj)

    def __str__(self):
        t = ", ".join([str(i) for i in self.objects])
        return f"<Group : {t} >"

    def __repr__(self):
        return self.__str__()

    def filter(self, func) -> 'Group':
        d = {k: v for k, v in self.items() if func(v)}
        return d


    # getters        
    def __getitem__(self, key):
        if isinstance(key, int):
            return self.__objects[list(self.__objects.keys())[key]]
        return self.__objects[key]

    def __getattr__(self, name):
        return self.__objects[name]
    
    def values(self, no_parent=False):
        if no_parent and self.__parent:
            return [v for k, v in self.__objects.items() if v != self.__parent]
        
        return self.__objects.values()
    
    def items(self, no_parent=False):
        if no_parent and self.__parent:
            return {k: v for k, v in self.__objects.items() if v != self.__parent}
        
        return self.__objects.items()

    def keys(self, no_parent=False):
        if no_parent and self.__parent:
            return [k for k, v in self.__objects.items() if v != self.__parent]
    
        return self.__objects.keys()
    
    # delete item
    def __delitem__(self, key):
        self.__objects.pop(key)
        self.__dict__.pop(key)
        
    def __delattr__(self, key):
        self.__objects.pop(key)
        self.__dict__.pop(key)
    
    # contains
    def __contains__(self, key):
        return key in self.__objects
    
    # iterables
    def __iter__(self):
        return iter(self.values())
    
    def __next__(self):
        return next(self.values())

    def __len__(self):
        return len(self.__objects)
 
class Line:
    r"""
	#### Line
 
	#### Parameters
	- `start`: start of the line `[x, y]`
	- `end`: end of the line `[x, y]`
	- width : width of the line `int`
	- `color`: color of the line `"white" or (R, G, B)`
	- `stroke`: stroke of the line `int`
	
    """
    __slots__ = ("start", "end", "width", "color", "window")
    
    def __init__(self, start: Pos, end: Pos, width: int = 5, color: Color = "white"):
        self.start = start
        self.end = end
        self.width = width
        self.color = resolve_color(color)
        self.window = get_window()

    def draw(self):
        pg.draw.line(self.window.surface, self.color,
                     self.start, self.end, self.width)

    def _get_collision_box(self):
        # collision is checked at line end
        return (self.end, self.end + (self.width, 0), self.end + (0, self.width), self.end + (self.width, self.width))



import pygame as pg, os

FONTS_PATH = "ezsgame/assets/fonts"

class FontFamily:
    def __init__(self, folder_name:str, main_font:str, is_variation:bool = False):
        
        """
        if is_variation is True then main_font is the file name of the font, not just the name of the font
        """
        # is variation font
        if is_variation:
            self.font_file = f"{FONTS_PATH}/{folder_name}/{main_font}"
        
        
        # if font family  get main font and variations
        else:

            # get variations                
            for filename in os.listdir(FONTS_PATH + "/" + folder_name):
                
                if filename.endswith(".ttf"):
                    
                    font_name = filename.replace(f"{folder_name}-", "")
                    font_name = font_name.replace(".ttf", "")
                    
                     # set main font
                    if font_name == main_font:
                        self.font_file = f"{FONTS_PATH}/{folder_name}/{filename}"
                        continue            

                    # set variations
                    setattr(self, font_name.lower(), FontFamily(folder_name, filename, True))
            
    def get_font(self, font_size) -> pg.font.Font:
        return pg.font.Font(self.font_file, font_size)
    

# default fonts  
class Fonts:
    RobotoMono: FontFamily = FontFamily("RobotoMono", "Regular")
    OpenSans = FontFamily("OpenSans", "Regular")


from typing import List

from ezsgame.types import Pos
from .global_data import get_window
import pygame as pg

def outline(obj, color="red", stroke:int=1, size:int=1.5, border_radius:list = [0,0,0,0]):
    r'''
    #### Draws an outline of the object
    - `color` : the color of the outline `White`or `(R,G,B)`
    - `stroke` : the width of the outline `int`
    - `size` : the size  multiplier of the outline (object size * outline size) `int`
    - `border_radius` : the radius of the corners of the outline `[top_left, top_right, bottom_left, bottom_right]`

    #### Note: Automaticly draws the outline when declared
    '''
    
    obj_size = obj.get_size()
    obj_pos = obj.get_pos()

    size = (obj_size[0] * size, obj_size[1] * size)
    pos = (obj_pos[0] - (size[0] - obj_size[0]) / 2, obj_pos[1] - (size[1] - obj_size[1]) / 2)

    pg.draw.rect(get_window().surface, color, [*pos, *size], stroke, *border_radius)
    
def is_out(obj):
    r'''
    Return True if objects is out of bounds and direction of that bound (top, bottom, right, left)
    -> [bool, "direction"]
    '''
    screen = get_window()
    
    if obj.pos[0] + obj.size[0] < 0 or obj.pos[0] - obj.size[0]/4 > screen.size[0]:
        return True, "left" if obj.pos[0] + obj.size[0]/2 <= 0 else "right"
    elif obj.pos[1] + obj.size[1] < 0 or obj.pos[1] - obj.size[1]/4 > screen.size[1]:
        return True, "top" if obj.pos[1] + obj.size[1] <= 0 else "bottom"     
    else:
        return False, None
        
def is_colliding(obj1, obj2, draw_collision_box=False):
    r'''
    #### returns True if the object is colliding with another object, False if not
    - obj1 : object -> first object
    - obj2 : object -> second object
    - draw_collision_box : bool -> if True, draw collision box
    '''
    
    if draw_collision_box:
        outline(obj1, size=1.1, stroke=2)
        outline(obj2, size=1.1, stroke=2)

    for i in obj1._get_collision_box():    
        if (i[0] >= obj2.pos[0] and i[0] <= obj2.pos[0]+obj2.size[0]) and (i[1] >= obj2.pos[1] and i[1] <= obj2.pos[1] + obj2.size[1]):
            return True
                
        
    return False
        
        
def div(axis : str, q : int, size : float = None) -> List[List[float]]:
		r'''
		#### Returns a list of division points of the screen in the given axis
		
		- `axis` : axis to divide the screen in (`x` or `y`)
		- `q` : number of divisions
		- `size` : Size of where to divide the screen, works as a delimiter (Optional)
        - `pos_prefix` : Position to add to the division points (Optional)
		'''
		
		divs = []
		if axis == "x":
			step = size[0] / q
			
			for i in range(q):
				divs.append([round(i * step, 1), round((i + 1) * step, 1) ])
				
				# if overflows 
				if divs[-1][1] > size[0]:
					break
				
		elif axis == "y":
			step = size[1] / q

			for i in range(q):
				divs.append([round(i * step, 1), round((i + 1) * step, 1)])
		
				# if overflows
				if divs[-1][1] > size[1]:
					break
				
		return divs

# Positioning functions
def center_of(obj) -> Pos:
    r'''
    #### Returns the center postion of the object
    - obj : object
    '''
    return Pos(obj.pos[0] + obj.size[0]/2, obj.pos[1] + obj.size[1]/2)


def center_at(obj, parent, x: bool = True, y: bool = True):
    r'''
    #### Centers an object in the parent object
    - obj : object -> object to center
    - parent : object -> parent object
    - x -> if True, center x-axis
    - y -> if True, center y-axis
    '''
    
    if x:
        obj.pos[0] = parent.pos[0] + (parent.size[0] - obj.size[0]) / 2
    if y:
        obj.pos[1] = parent.pos[1] + (parent.size[1] - obj.size[1]) / 2


def is_hovering(obj):
    """
    #### Returns True if the mouse is hovering over the object, False if not
    - obj : object -> object
    """

    box = obj._get_collision_box()
    
    mouse_pos = pg.mouse.get_pos()
    
    if mouse_pos[0] >= box[0][0] and mouse_pos[0] <= box[1][0] and mouse_pos[1] >= box[0][1] and mouse_pos[1] <= box[1][1]:
        return True
    else:
        return False





from typing import Iterable
from path import Path
from ezsgame.styles.units import Measure
from ezsgame.types import Pos, Size


class Tile:
    def __init__(
            self,
            sprite: Path | str,
            pos: Pos | Iterable[Measure],
            size: Measure # Tiles are always square
    ):
        self.sprite = sprite
        self.pos = pos
        self.size = size
        self.rect = (self.pos, self.size)
    
    
    def __repr__(self):
        return f"Tile(sprite={self.sprite}, pos={self.pos}, size={self.size})"
    
    def __str__(self):
        return f"Tile(sprite={self.sprite}, pos={self.pos}, size={self.size})"
    

class TileMap:
    def __init__(
            self,
            tiles: list[Tile],
            pos: Pos | Iterable[Measure],
            size: Size | Iterable[Measure]
    ):
        self.tiles = tiles
        self.pos = pos
        self.size = size
        self.rect = (self.pos, self.size)
    
    
    def __repr__(self):
        return f"TileMap(tiles={self.tiles}, pos={self.pos}, size={self.size})"
    
    def __str__(self):
        return f"TileMap(tiles={self.tiles}, pos={self.pos}, size={self.size})"


"""
Module for scenes manipulation and creation
"""

from abc import ABC, abstractmethod
from typing import FrozenSet

from .global_data import get_window

class Scene(ABC):
    """
    Abstract class for scenes

    ### Init
    Note: the minimum required for a scene is to have an `init`, `update` and `draw` methods
    - `name`: Name of the scene (defaults to class name without `_scene`) Example: `main_scene` -> `main`
    - `shadow_update`: If True, the scene will be updated even if it's not the current scene
    - `shadow_draw`: If True, the scene will be drawn even if it's not the current scene
    """

    def __init__(
        self, name: str = None, shadow_update: bool = False, shadow_draw: bool = False
    ):
        self.name: str = name or self.__class__.__name__.replace("_scene", "")
        self.shadow_update: bool = shadow_update
        self.shadow_draw: bool = shadow_draw

        self.manager: SceneManager = None


    @abstractmethod
    def init(self) -> None:
        """
        Initialize the scene
        (Happens before main loop)
        """

    @abstractmethod
    def update(self) -> None:
        """
        Update the scene
        (Happens before draw)
        """

    @abstractmethod
    def draw(self) -> None:
        """
        Draw the scene
        """

    def on_switch(self) -> None:
        """
        Called when the scene is switched to
        """
        pass

    def on_switch_out(self) -> None:
        """
        Called when the scene is switched from
        """
        pass

    def switch_to(self, scene_name: str) -> None:
        """
        Switch to a scene
        """
        if self.manager:
            self.manager.switch_to(scene_name)
        else:
            raise Exception("Scene manager not found")


class SceneManager:
    """
    Class for managing scenes

    ### Init
    - `scenes`: Scenes to be managed
    - `main_scene`: Name of the main scene (defaults to `main`)
    - `lazy_load`: If True, scenes will be initialized (init method) only when switched to not when the scene manager is initialized (Main scene will always be initialized)
    """

    def __init__(self, *scenes: Scene, main_scene: str = "main", lazy_load: bool = False):

        self.window = get_window()

        self.lazy_load: bool = lazy_load
        self.scenes: dict = {scene.name: scene for scene in scenes}
        
        try:
            self.current_scene: Scene = self.scenes[main_scene]
            self.current_scene.init()
            self.current_scene.on_switch()
        except KeyError:
            raise Exception(f"Main scene '{main_scene}' not found in scenes <{self.scenes.keys()}>")


        # Reference to scenes that have shadow_update
        self.has_shadow_update: FrozenSet[str] = frozenset(
            [scene.name for scene in scenes if getattr(scene, "shadow_update", False)]
        )

        # Reference to scenes that have shadow_draw
        self.has_shadow_draw: FrozenSet[str] = frozenset(
            [scene.name for scene in scenes if getattr(scene, "shadow_draw", False)]
        )

        for scene in scenes:
            scene.manager = self

        if not lazy_load:
            for scene in scenes:
                scene.init()

    def switch_to(self, scene_name: str) -> None:
        """
        Switch to a scene
        """
        if scene_name in self.scenes.keys():
            if self.current_scene:
                self.current_scene.on_switch_out()
                self.window.fill() # Clear window

            self.current_scene = self.scenes[scene_name]

            if self.lazy_load:
                self.current_scene.init()

            self.current_scene.on_switch()
        else:
            raise Exception(f"Scene '{scene_name}' not found")

    def update(self) -> None:
        """
        Update sthe current scene (and scenes that have shadow_update)
        """
        if self.current_scene:
            self.current_scene.update()

            # Update scenes that have shadow_update
            map(
                lambda name: self.scenes[name].update(),
                self.has_shadow_update - {self.current_scene.name},
            )

    def draw(self) -> None:
        """
        Draws the current scene (and scenes that have shadow_draw)
        """
        if self.current_scene:
            self.current_scene.draw()

            # Draw scenes that have shadow_draw
            map(
                lambda name: self.scenes[name].draw(),
                self.has_shadow_draw - {self.current_scene.name},
            )

    def __del__(self) -> None:
        """
        Exits all scenes
        """
        for scene in self.scenes.values():
            del scene # Calls exit

    # Utility methods ----------------------------------------------
    def __getitem__(self, scene_name: str) -> Scene:
        """
        Returns a scene
        """
        return self.scenes[scene_name]



from typing import Callable, List
import pygame as pg
from enum import Enum
    
class TimeUnit(Enum):
    milliseconds = 1
    seconds = 1000
    minutes = 60000
    hours = 3600000
    days = 86400000

# Condence the time units into one class 
class Milliseconds:
    def __init__(self, time: float):
        self.time = time

    @classmethod
    def from_unit(cls, unit: TimeUnit, time: float):
        return cls(time * unit.value)


class Interval:
    __slots__ = "time", "callback", "name", "last_call"
    
    def __init__(self, time: float, callback: Callable, last_call: float, name: str = "Default"):
        self.time = time
        self.callback = callback
        self.name = name
        self.last_call = last_call

class TimeHandler:
    r'''
    - Handles the time events
    '''

    intervals: List[Interval] = []
    to_remove: List[str] = []
    to_add: List[Interval] = []

    def add(call_time: int, callback, name: str = "Default"):
        r'''
        #### Adds a `interval` that will be called every `time` milliseconds
        - `name` : name of the event 
        - `time` : amount of time in milliseconds that the event will be called after
        - `callback` : function to be called when the event is triggered 
        '''

        name = f"{len(TimeHandler.intervals)}.{call_time}" if name == "Default" else name

        TimeHandler.to_add.append(Interval(call_time, callback, 0, name)) 

    def remove(name: str):
        r'''
        #### Removes an `interval` from the event list so it won't be called anymore
        - `name` : name of the event to be removed 
        '''
        TimeHandler.to_remove.append(name)

    def check():
        r'''
        #### Manages the time events
        '''
        # removing intervals
        for name in TimeHandler.to_remove:

            if name in TimeHandler.intervals:
                del TimeHandler.intervals[name]

        TimeHandler.to_remove.clear()

        # adding intervals
        TimeHandler.intervals.extend(TimeHandler.to_add)
       
        TimeHandler.to_add.clear()

        # Checking  Intervals
        
        current_time = pg.time.get_ticks()
        for interval in TimeHandler.intervals:
            
            if current_time - interval.last_call >= interval.time:
                interval.callback()
                interval.last_call = pg.time.get_ticks()

# time decorators  ------------------------------------------------------------
def add_interval(time: Milliseconds | int, name: str = "Default") -> Callable:
    r'''    
    - Adds an `interval` to the time handler, calls the function every `time` milliseconds
    - `time` : amount of time in milliseconds that the event will be called after
    - `name` : name of the interval (Optional)
    '''

    if isinstance(time, Milliseconds):
        time = time.time

    def wrapper(func):
        TimeHandler.add(time, func, name)
        return func

    return wrapper

def remove_interval(name: str) -> None:
    r'''
    #### Removes an `interval` from the time handler
    - `name` : name of the interval
    '''
    TimeHandler.remove(name)


